"""Test various mathematical formulae used in GALUMPH.

This uses symbolic algebra to verify various simplified formulae.
"""

# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import pytest

sympy = pytest.importorskip('sympy')


def iLMKP(L, M, K, P, check=True):
    """Index into a packed DLMKP matrix."""
    if check:
        assert L >= M
        assert L >= K
        assert M >= 0
        assert K >= M
        assert (L+K) >= P
        assert (L-K) <= P
        assert (L+K+P) % 2 == 0
    iP = (P - (L-K)) // 2
    iK0 = (K*(K+1) - M*(M+1)) // 2
    iM00 = (3*L*(L+3) + 7 - M*M) * M // 6
    iL000 = L*(L+1)*(L+1)*(L+2) // 12
    return iL000 + iM00 + iK0 + iP


def assert_sympy_equal(a, b):
    """Assert that two sympy expressions are equal."""
    assert sympy.simplify(a - b) == 0


def test_packed_sizes_offsets():
    """Check formulae for sizes of and offsets into packed dlmkp matrices."""
    def psize(K):
        """Number of P points for a given K (where K <= L)."""
        return K + 1

    def pindex(L, K, P):
        """Number of points between (L,M,K,Pmin) and (L,M,K,P)."""
        return (P - (L-K)) / 2

    def ksize(L, M):
        """Number of (K, P) points for given (L, M)."""
        K = sympy.symbols("k", integer=True)
        return sympy.summation(psize(K), (K, M, L))

    def iKP_local_ip(M, K, iP):
        """Number of points between (L,M,Kmin,Pmin) and (K,P)."""
        Kx = sympy.symbols("k'", integer=True)
        return sympy.summation(psize(Kx), (Kx, M, K-1)) + iP

    def iKP_local(L, M, K, P):
        """Number of points between (L,M,Kmin,Pmin) and (K,P)."""
        return iKP_local_ip(M, K, pindex(L, K, P))

    def msize(L):
        """Number of points for a given L."""
        M = sympy.symbols("M", integer=True)
        return sympy.summation(ksize(L, M), (M, 0, L))

    def lsize(LMAX):
        """Total number of points for L <= LMAX."""
        L = sympy.symbols("L", integer=True)
        return sympy.summation(msize(L), (L, 0, LMAX))

    def iL000(L):
        """Index of (L,0,0,Pmin)."""
        Lx = sympy.symbols("L'", integer=True)
        return sympy.summation(msize(Lx), (Lx, 0, L-1))

    def iLMoff(L, M):
        """Difference between (L,0,0,Pmin) and (L,M,Kmin,Pmin)."""
        Mx = sympy.symbols("M'", integer=True)
        return sympy.summation(ksize(L, Mx), (Mx, 0, M-1))

    def iLM00(L, M):
        """Index of (L,M,0,Pmin)."""
        return iL000(L) + iLMoff(L, M)

    # Put these checks in internal sub-functions so that the variables
    # are not visible in the scope of the helper sub-functions above
    def check_sizes():
        """Check that sizes are one greater than the index of the last value."""
        L, M, K = sympy.symbols("L M, K", integer=True)
        assert_sympy_equal(
            psize(K) - 1,
            pindex(L, K, L+K),
        )
        assert_sympy_equal(
            ksize(L, M) - 1,
            iKP_local(L, M, L, 2*L),
        )
        assert_sympy_equal(
            lsize(L) - 1,
            iLM00(L, L) + iKP_local(L, L, L, 2*L),
        )

    def check_formulae():
        L, M, K, P, ip, LMAX = sympy.symbols("L M, K, P, i_P, LMAX", integer=True)
        assert_sympy_equal(
            ((L+1) * (L+2) - M * (M+1)) / 2,
            ksize(L, M),
        )
        assert_sympy_equal(
            (K * (K+1) - M * (M+1)) / 2 + ip,
            iKP_local_ip(M, K, ip),
        )
        assert_sympy_equal(
            (LMAX+1) * (LMAX+2)**2 * (LMAX+3) / 12,
            lsize(LMAX),
        )
        assert_sympy_equal(
            L * (L+1) * (L+1) * (L+2) / 12,
            iL000(L),
        )
        assert_sympy_equal(
            (3 * L * (L+3) + 7 - M * M) * M / 6,
            iLMoff(L, M),
        )

    check_sizes()
    check_formulae()



# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl
import pyopencl.array as cl_array

from .almarray import AlmArray
from .sphericalbessel import SphericalBessel
from .util import ClProgram, LMAXMixin, check_array, with_some_queue
from .gaunt import shzmat


class PackedLMKPMixin(LMAXMixin):
    """Utility functions for packed [L,M,K,P] arrays."""

    @property
    def NLMKP(self):
        """Total number of points in the packed array."""
        return (self.LMAX+1) * (self.LMAX+2)**2 * (self.LMAX+3) // 12


class AlmShiftZ(PackedLMKPMixin, ClProgram):
    """OpenCL kernel for translating ALM arrays along the Z axis."""

    def __init__(self, LMAX, NSWORK, s, dlmkp=None, context=None):
        """Compile the kernel."""
        self.context = context
        self.LMAX = LMAX
        self.NS = len(s)
        self.s = s
        # TODO work out a default value for NSWORK so the user does not have to set it
        self.NSWORK = NSWORK
        assert LMAX >= 0
        assert self.NS > 0
        assert self.NSWORK > 0
        assert self.NS % self.NSWORK == 0

        if dlmkp is None:
            dlmkp = shzmat(LMAX)
        check_array(dlmkp, (self.NLMKP,), exact=False)

        super(AlmShiftZ, self).__init__(
            filename='translate-alm.cl',
            options=["-DLMAX1=%d" % self.LMAX1, "-DNSWORK=%d" % NSWORK],
            context=context)

        # Copy the dlmkp array to the device if it is not already a device array
        if isinstance(dlmkp, cl_array.Array):
            assert dlmkp.context == self.context
            assert dlmkp.dtype == np.dtype('f4')
            self.dlmkp = dlmkp
        else:
            dlmkp = np.asarray(dlmkp, dtype='f4')
            with cl.CommandQueue(self.context) as queue:
                self.dlmkp = cl_array.to_device(queue, dlmkp)

        self.sphbes = SphericalBessel(
            2*LMAX, s,
            # TODO the SphericalBessel kernel can use a larger work size
            worksize=NSWORK,
            natwork=1,
            context=self.context,
        )

    @with_some_queue
    def __call__(self, alm, u, queue=None):
        """Shift Alm by u along the z axis and return as Blm."""
        assert alm.context == self.context
        assert alm.LMAX == self.LMAX
        assert alm.NS == self.NS
        assert queue is not None
        if (u < 0):
            # TODO implement negative shifts
            raise NotImplementedError('Negative z-shifts not yet supported.')
        LMAX = self.LMAX
        LMAX1 = self.LMAX1
        NS = self.NS
        NSWORK = self.NSWORK

        ev_j, jp = self.sphbes(u, queue=queue)

        blm = AlmArray(queue, LMAX, NS)
        alm_local = cl.LocalMemory(LMAX1*NSWORK * 8)
        blm_local = cl.LocalMemory(LMAX1*NSWORK * 8)
        dlmkp_local = cl.LocalMemory(LMAX1*LMAX1 * 4)
        jp_local = cl.LocalMemory((2*LMAX + 1)*NSWORK * 4)

        ev = self.program.translate_alm(queue,
                                        (LMAX1, LMAX1, NS),
                                        (LMAX1, 1, NSWORK),
                                        alm.data,
                                        blm.data,
                                        self.dlmkp.data,
                                        jp.data,
                                        alm_local,
                                        blm_local,
                                        dlmkp_local,
                                        jp_local,
                                        wait_for=(ev_j,))
        return ev, blm



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2017-2019 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "European Molecular Biology Laboratory (EMBL)"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl
from pyopencl import array as cl_array

from .util import ClProgram, LMAXMixin, optional_out_array


class SphericalBessel(LMAXMixin, ClProgram):
    """Calculate an array of spherical Bessel functions

    Calculates the spherical Bessel function of the first kind j_L(r*s)
    for all L up to and including LMAX.
    The s array is assumed to be the same for all invocations and is passed
    to the __init__ function; r is the only parameter which varies between
    invocations."""

    def __init__(self, LMAX, s, worksize=64, natwork=16, context=None):
        """Set up the arrays, kernels etc for the spherical Bessel function

        s must be a one-dimensional array whose size is a multiple of
        the worksize parameter. LMAX should be a positive integer."""
        assert LMAX >= 0
        self.LMAX = LMAX
        s = np.asanyarray(s)
        NS, = s.shape
        assert NS > 0
        assert np.all(s >= 0)
        self.NS = NS
        assert worksize > 0
        assert NS % worksize == 0
        self.worksize = worksize
        assert natwork > 0
        self.natwork = natwork
        super(SphericalBessel, self).__init__(
            filename='spherical-bessel.cl',
            options=["-DWORKSIZE=%d" % worksize],
            context=context)
        context = self.context

        self.s_host = np.array(s, dtype=np.float32)
        with cl.CommandQueue(context) as queue:
            self.s_dev = cl_array.to_device(queue, self.s_host)

    @optional_out_array(shape=('LMAX1', 'NS'), dtype='f4')
    def __call__(self, r, queue=None, out=None):
        """Calculate the spherical Bessel function j_L(r*s) for L<=LMAX

        The calculation is done asynchronously, the returned OpenCL
        event object can be used to wait for the result."""
        assert r >= 0
        NS = self.NS
        worksize = self.worksize

        ev = self.program.jsph(queue,
                               (NS,),
                               (worksize,),
                               np.uint32(self.LMAX),
                               np.float32(r),
                               self.s_dev.data,
                               out.data)

        return ev, out

    @optional_out_array(shape=('natwork', 'LMAX1', 'NS'), dtype='f4')
    def batch(self, r, queue=None, out=None):
        """Batch spherical Bessel function for an array of r values."""
        r = np.array(r, dtype=np.float32)
        assert np.all(r >= 0)
        NS = self.NS
        worksize = self.worksize
        nAt = np.size(r)
        assert nAt > 0
        assert nAt <= self.natwork

        r_dev = cl_array.to_device(queue, r)
        ev = self.program.jsph_batch(queue,
                                     (NS, nAt),
                                     (worksize, 1),
                                     np.uint32(self.LMAX),
                                     r_dev.data,
                                     self.s_dev.data,
                                     out.data)

        return ev, out

    @optional_out_array(shape=('LMAX1', 'NS'), dtype='f4')
    def integral(self, r, queue=None, out=None):
        """Calculate the spherical Bessel integral \\Integral{0}{r} j_L(r'*s) dr' for L<=LMAX

        The calculation is done asynchronously, the returned OpenCL
        event object can be used to wait for the result.
        """
        assert r >= 0
        NS = self.NS
        LMAX = self.LMAX
        LMAX1 = self.LMAX1
        s_dev = self.s_dev
        worksize = self.worksize

        ev1 = self.program.jsph(queue,
                                (NS, 1),
                                (worksize, 1),
                                np.uint32(LMAX),
                                np.float32(r),
                                s_dev.data,
                                out.data)

        localJL = cl.LocalMemory(4*NS)
        ev = self.program.jsph_integral(queue,
                                        (worksize, LMAX1),
                                        (worksize, 1),
                                        np.uint32(NS),
                                        np.float32(r),
                                        s_dev.data,
                                        out.data,
                                        localJL,
                                        wait_for=(ev1,))

        return ev, out



# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import given, settings
import hypothesis.strategies as st
import numpy as np
import numpy.testing as npt

from galumph.atomicscattering import AtomicScattering
from galumph.rotalm import RotateAlm

# Import test strategies from test_atomic_scattering
from test_atomic_scattering import alm_parameters, atom_xyzf


@given(
    LMAX=st.integers(min_value=0, max_value=2),
    beta=st.floats(min_value=-np.pi, max_value=np.pi),
)
def test_beta_rotation_matrix_analytical(cl_context, LMAX, beta):
    """Compare early elements of the dlmm matrix with analytical solutions."""
    kernel = RotateAlm(context=cl_context, LMAX=LMAX, NSWORK=8)
    dlmm_dev = kernel.y_rotation_matrix(beta)
    dlmm = dlmm_dev.get()
    # N.B. in the Wikipedia article, the M2 comes before the M1
    npt.assert_approx_equal(dlmm[0], 1)
    if LMAX == 0:
        return
    cosbeta = np.cos(beta)
    sinbeta = np.sin(beta)
    sqrt2 = np.sqrt(2)
    analytical_dlmm_1 = np.array([[
        0,  # -1, 0  # TODO remove
        (1 - cosbeta) / 2,  # -1, 1
    ], [
        cosbeta,  # 0, 0
        -sinbeta / sqrt2,  # 0, 1
    ], [
        0,  # 1, 0  # TODO remove
        (1 + cosbeta) / 2,  # 1, 1
    ]])
    analytical_dlmm_1[0, 0] = analytical_dlmm_1[1, 1]
    analytical_dlmm_1[2, 0] = -analytical_dlmm_1[1, 1]

    dlmm_1 = dlmm[1:7].reshape((3, 2))
    npt.assert_allclose(dlmm_1, analytical_dlmm_1, rtol=1e-4, atol=1e-6)
    if LMAX == 1:
        return
    cos2beta = cosbeta**2
    sqrt38 = np.sqrt(3/8)

    analytical_dlmm_2 = np.array([[
        0,  # -2, 0  # TODO remove
        0,  # -2, 1  # TODO remove
        (1 - cosbeta)**2 / 4,  # -2, 2
    ], [
        0,  # -1, 0  # TODO remove
        (-2 * cos2beta + cosbeta + 1) / 2,  # -1, 1
        -sinbeta * (1 - cosbeta) / 2,  # -1, 2
    ], [
        (3 * cos2beta - 1) / 2,  # 0, 0
        -sqrt38 * np.sin(2*beta),  # 0, 1
        sqrt38 * sinbeta**2,  # 0, 2
    ], [
        0,  # 1, 0  # TODO remove
        (2 * cos2beta + cosbeta - 1) / 2,  # 1, 1
        -sinbeta * (1 + cosbeta) / 2,  # 1, 2
    ], [
        0,  # 2, 0  # TODO remove
        0,  # 2, 1  # TODO remove
        (1 + cosbeta)**2 / 4,  # 2, 2
    ]])
    # Apply symmetry - TODO pack the array more
    for M1 in range(-2, 3):
        iM1 = 2 + M1
        for M2 in range(3):
            if M1 > M2:
                # d[M2,M1] = -1^(M2-M1) * d[M1,M2]
                sign = 1 - 2 * ((M1 - M2) % 2)
                analytical_dlmm_2[iM1, M2] = sign * analytical_dlmm_2[2+M2, M1]
            elif -M1 > M2:
                # d[-M2,-M1] = d[M1,M2]
                analytical_dlmm_2[iM1, M2] = analytical_dlmm_2[2-M2, -M1]

    dlmm_2 = dlmm[7:].reshape((5, 3))
    npt.assert_allclose(dlmm_2, analytical_dlmm_2, rtol=1e-4, atol=1e-6)


@settings(max_examples=10)
@given(
    almparams=alm_parameters(worksize=st.integers(min_value=1, max_value=16)),
    xyz_ff=atom_xyzf(xyz_elements=st.floats(min_value=-5, max_value=5, width=32)),
    alpha=st.floats(min_value=-np.pi, max_value=np.pi),
    beta=st.floats(min_value=-np.pi, max_value=np.pi),
    gamma=st.floats(min_value=-np.pi, max_value=np.pi),
)
def test_atomic_scattering_rotated(cl_context, cl_queue, almparams, xyz_ff,
                                   alpha, beta, gamma):
    """Compare rotating the Alm versus rotating the atomic positions."""
    scattering_kernel = AtomicScattering(context=cl_context, **almparams)
    rotation_kernel = RotateAlm(context=cl_context,
                                LMAX=almparams['LMAX'],
                                NSWORK=almparams['worksize'],)
    xyz, ff = xyz_ff

    alm_orig = scattering_kernel(xyz, ff)

    ev, alm_rotalm = rotation_kernel(alm_orig, alpha, beta, gamma,
                                     queue=cl_queue)

    rotate_alpha = np.array([[np.cos(alpha), np.sin(alpha), 0],
                             [-np.sin(alpha), np.cos(alpha), 0],
                             [0, 0, 1]])
    rotate_beta = np.array([[np.cos(beta), 0, -np.sin(beta)],
                            [0, 1, 0],
                            [np.sin(beta), 0, np.cos(beta)]])
    rotate_gamma = np.array([[np.cos(gamma), np.sin(gamma), 0],
                             [-np.sin(gamma), np.cos(gamma), 0],
                             [0, 0, 1]])
    rotation = np.dot(rotate_alpha, np.dot(rotate_beta, rotate_gamma))
    xyzrot = np.dot(xyz, rotation)
    alm_rotxyz = scattering_kernel(xyzrot, ff)

    ev.wait()
    np_alm_rotalm = alm_rotalm.get(unpack=True)
    # M==0 terms should be real
    npt.assert_array_almost_equal(np_alm_rotalm[:, 0, :].imag, 0, decimal=4)

    tolerance_factor = max(1, np.max(xyz) * np.max(ff) * np.max(almparams['s']))
    np_alm_rotxyz = alm_rotxyz.get(unpack=True)
    npt.assert_allclose(np_alm_rotalm, np_alm_rotxyz,
                        rtol=1e-3 * tolerance_factor,
                        atol=1e-5 * tolerance_factor)



# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

import functools

import numpy as np
from pkg_resources import resource_string
import pyopencl as cl
import pyopencl.array as cl_array


def with_some_queue(f):
    """Decorator which creates a queue if none is passed in the keywords.

    The wrapped function must be a method on an object with a `.context`
    attribute and must return an event and a value. If a queue is passed,
    the (event, value) tuple will be returned as-is (asynchronous). If no
    queue is passed, a new queue will be created and the wrapper waits for
    the event to complete before returning a value.

    """
    @functools.wraps(f)
    def wrapper(instance, *args,
                # queue=None,  # # Python 2 does not like this
                **kwargs):
        queue = kwargs.pop('queue', None)
        assert isinstance(instance.context, cl.Context)
        if queue is None:
            with cl.CommandQueue(instance.context) as queue:
                ev, value = f(instance, *args, queue=queue, **kwargs)
                ev.wait()
                return value
        else:
            assert instance.context == queue.context
            ev, value = f(instance, *args, queue=queue, **kwargs)
            return ev, value
    return wrapper


def check_array(array, shape, dtype=None, exact=True):
    """Check that the array has the correct shape.

    If exact is False, allow the leading dimension to be larger than the
    expected shape.
    """
    if dtype is not None:
        dtype = np.dtype(dtype)
        if array.dtype != dtype:
            raise TypeError('Array should have dtype %s, got %s' %
                            (dtype, array.dtype))

    if array.shape == shape:
        return
    elif ((not exact) and
          (array.shape[1:] == shape[1:]) and
          (array.shape[0] >= shape[0])):
        return
    else:
        raise ValueError('Array shape %r does not match expected shape %s' %
                         (array.shape, shape))


def optional_out_array(f=None, shape=None, dtype=None, **check_kwargs):
    """Decorator creating an output array if none is passed in the keywords.

    Entries in the shape tuple can either be integers, in which case they are
    used as-is, or strings, in which case they are looked up as attributes
    on the object of the wrapped method. Additional keyword arguments are
    passed to the `check_array()` function to check the shape of the passed
    array if one is passed.

    This includes the with_some_queue functionality as creating an array
    requires a queue.

    """
    shape = tuple(shape)
    dtype = np.dtype(dtype)
    if f is None:
        return functools.partial(optional_out_array,
                                 shape=shape, dtype=dtype, **check_kwargs)

    @with_some_queue
    @functools.wraps(f)
    def wrapper(instance, *args,
                # queue=None, out=None,  # Python 2 does not like this
                **kwargs):
        queue = kwargs.pop('queue')
        out = kwargs.pop('out', None)
        required_shape = []
        for item in shape:
            if isinstance(item, str):
                item = getattr(instance, item)
            required_shape.append(item)
        required_shape = tuple(required_shape)
        if out is None:
            out = cl_array.empty(queue, required_shape, dtype)
        check_array(out, required_shape, dtype=dtype, **check_kwargs)
        assert instance.context == out.context
        assert queue == out.queue
        ev, value = f(instance, *args, queue=queue, out=out, **kwargs)
        # The returned array should be the same as the `out` array
        assert value is out
        return ev, value
    return wrapper


class ClProgram(object):
    """Base class for OpenCL kernel programs."""

    def __init__(self, filename, options=(), context=None):
        """Build the CL program from the source file.

        Create a context using `cl.create_some_context()` if None is passed.
        """
        if context is None:
            context = cl.create_some_context()
        self.context = context
        # The path is internal to pkg_resources i.e. always '/' separators
        filepath = '/'.join(('cl-src', filename))
        program_src = resource_string(__name__, filepath)
        if not isinstance(program_src, str):  # Python 3
            program_src = str(program_src, 'ascii')
        self.program = cl.Program(context, program_src)
        self.program.build(options=options)


class LMAXMixin(object):
    """Super simple mixin which adds an LMAX1 attribute with value LMAX+1."""

    @property
    def LMAX1(self):
        return self.LMAX + 1


class PackedLMMixin(LMAXMixin):
    """Mixin class with methods for arrays with packed [L,M] dimensions.

    This version is for arrays only containing non-negative M values.

    If the LMAX attribute is set on the class, LMAX1 and NLM are available
    as properties.
    """

    @staticmethod
    def indexLM(L, M=0):
        """Index into a packed [L,M] array."""
        assert M >= 0
        assert L >= M
        return L * (L + 1) // 2 + M

    @property
    def NLM(self):
        return self.indexLM(self.LMAX1, 0)



# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl

from .almarray import AlmArray
from .util import ClProgram, PackedLMMixin, optional_out_array, with_some_queue


def _restrict_angle(theta):
    """Restrict an angle to the range (-pi, pi)."""
    if abs(theta) <= np.pi:
        return theta
    return (theta - np.pi) % (2 * np.pi) - np.pi


class PackedLM1M2Mixin(PackedLMMixin):
    """Mixin with utility methods for packed [L, M1, M2] arrays.

    0 <= L <= LMAX
    -L <= M1 <= L
    0 <= M2 <= L
    """

    @staticmethod
    def indexM1M2(L, M1, M2):
        """Index into a [M1, M2] slice of a packed [L, M1, M2] array."""
        assert L >= 0
        assert abs(M1) <= L
        assert M2 >= 0
        assert M2 <= L
        return (L+M1) * (L+1) + M2

    @staticmethod
    def indexLM1M2(L, M1=0, M2=0):
        """Index into a packed [L, M1, M2] array."""
        assert L >= 0
        assert abs(M1) <= L
        assert M2 >= 0
        assert M2 <= L
        return L * (L+1) * (4*L-1) // 6 + (L+M1) * (L+1) + M2

    @property
    def LMAX21(self):
        """Number of elements between -LMAX and +LMAX."""
        return self.LMAX * 2 + 1

    @property
    def NM1M2(self):
        """Size of an [M1, M2] slice for L=LMAX."""
        return self.LMAX21 * self.LMAX1

    @property
    def shapeM1M2(self):
        """Shape of an [M1, M2] slice for L=LMAX."""
        return (self.LMAX21, self.LMAX1)

    @property
    def NLM1M2(self):
        """Number of [L, M1, M2] elements for L <= LMAX."""
        return self.indexLM1M2(self.LMAX1, -self.LMAX1, 0)


class RotateAlm(PackedLM1M2Mixin, ClProgram):
    """Python interface to the alm rotation kernels."""

    def __init__(self, LMAX, NSWORK, context=None):
        """Compile the kernel."""
        self.context = context
        self.LMAX = LMAX
        # TODO work out a default value for NSWORK so the user does not have to set it
        self.NSWORK = NSWORK
        super(RotateAlm, self).__init__(
            filename='rotate-alm.cl',
            options=["-DLMAX1=%d" % self.LMAX1, "-DNSWORK=%d" % NSWORK],
            context=context)

    @with_some_queue
    def __call__(self, alm, alpha, beta, gamma, queue=None):
        """Rotate Alm by the Euler angles and return as Blm."""
        assert alm.context == self.context
        assert alm.LMAX >= self.LMAX
        assert (alm.NS % self.NSWORK) == 0
        assert queue is not None
        alpha = _restrict_angle(alpha)
        beta = _restrict_angle(beta)
        gamma = _restrict_angle(gamma)
        LMAX = self.LMAX
        LMAX1 = self.LMAX1
        NS = alm.NS
        NSWORK = self.NSWORK

        ev_mat, dlmm = self.y_rotation_matrix(beta, queue=queue)

        blm = AlmArray(queue, LMAX, NS)
        eiM1alpha = cl.LocalMemory(8 * LMAX1)
        dlmm_local = cl.LocalMemory(4 * self.NM1M2)
        ev = self.program.rotate_alm(queue,
                                     (LMAX1, LMAX1, NS),
                                     (1, LMAX1, NSWORK),
                                     alm.data,
                                     blm.data,
                                     dlmm.data,
                                     np.float32(alpha),
                                     np.float32(gamma),
                                     eiM1alpha,
                                     dlmm_local,
                                     wait_for=(ev_mat,))
        return ev, blm

    @optional_out_array(shape=('NLM1M2',), dtype='f4')
    def y_rotation_matrix(self, beta, queue=None, out=None, wait_for=()):
        """Calculate the y rotation matrix (Wigner small d)."""
        assert queue is not None
        LMAX1 = self.LMAX1
        LMAX21 = self.LMAX21
        beta = _restrict_angle(beta)

        sterm_local = cl.LocalMemory(4 * LMAX1)
        jterm_local = cl.LocalMemory(4 * LMAX1)
        ev = self.program.y_rotation_matrix(queue,
                                            (LMAX21, LMAX1, LMAX1,),
                                            (1, 1, LMAX1),
                                            out.data,
                                            np.float32(beta),
                                            sterm_local,
                                            jterm_local,
                                            wait_for=wait_for)
        return ev, out



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2017-2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

import numpy as np
from pyopencl import array as cl_array

from .util import ClProgram, with_some_queue


class Intensity(ClProgram):
    """Functions for calculating the scattering intensity from Alm arrays."""

    def __init__(self, worksize, context=None):
        self.worksize = worksize
        super(Intensity, self).__init__('intensity.cl',
                                        options=['-DWORKSIZE=%d' % worksize],
                                        context=context)

    @with_some_queue
    def __call__(self, alm, queue=None):
        """Calculate the scattering intensity from the given ALM array.

        The intensity at each S point is the sum of the squared magnitude of
        all [L,M] indices at that S point.
        """
        assert alm.context == self.context
        assert (alm.NS % self.worksize) == 0
        assert queue is not None
        NS = alm.NS
        LMAX = alm.LMAX
        worksize = self.worksize

        int_dev = cl_array.empty(queue, (NS,), dtype='f4')

        ev = self.program.alm_intensity(queue,
                                        (NS,),
                                        (worksize,),
                                        np.uint32(LMAX),
                                        alm.data,
                                        int_dev.data)
        return ev, int_dev

    @with_some_queue
    def cross(self, alm, blm, queue=None):
        """Calculate the cross scattering between ALM and BLM."""
        assert alm.context == self.context
        assert blm.context == self.context
        assert (alm.NS % self.worksize) == 0
        assert blm.NS == alm.NS
        assert queue is not None
        NS = alm.NS
        LMAX = min(alm.LMAX, blm.LMAX)
        worksize = self.worksize

        int_dev = cl_array.empty(queue, (NS,), dtype='f4')

        ev = self.program.alm_cross(queue,
                                    (NS,),
                                    (worksize,),
                                    np.uint32(LMAX),
                                    alm.data,
                                    blm.data,
                                    int_dev.data)
        return ev, int_dev



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2017-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "European Molecular Biology Laboratory (EMBL)"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl
import pyopencl.array as cl_array

from .util import ClProgram, optional_out_array, PackedLMMixin


class SphericalHarmonic(PackedLMMixin, ClProgram):
    """Functor class giving spherical harmonics at a given theta, phi

    LMAX and the OpenCL work group size are set in the __init__ function
    The output will contain all values of L and all non-negative M up to
    and including LMAX, in a packed format

    theta and phi are passed to __call__ in the form of cos and sin values
    This is not only how they are used internally but also easier to calculate
    precisely from (x,y,z) coordinates"""

    def __init__(self, LMAX, natwork=16, context=None):
        """Set up all the necessary arrays, kernels etc"""
        if LMAX < 0:
            raise ValueError('LMAX < 0 is meaningless')
        elif LMAX < 3:
            # Using LMAX < 3 causes pocl to crash while compiling the
            # ylm_Mpart_local CL kernel - see issue #54
            raise NotImplementedError('LMAX < 3 is not supported')
        self.LMAX = LMAX
        assert natwork > 0
        self.natwork = natwork
        super(SphericalHarmonic, self).__init__(
            filename='spherical-harmonic.cl',
            options=["-DLMAX1=%d" % (self.LMAX1,)],
            context=context)

    @optional_out_array(shape=('NLM',), dtype='f4')
    def __call__(self, costheta, sintheta, queue=None, out=None):
        """Calculate all non-negative spherical harmonics up to LMAX.

        The calculation is done asynchronously, the returned OpenCL
        event object can be used to wait for the result.
        """
        assert -1 <= costheta
        assert costheta <= 1
        assert 0 <= sintheta  # 0 <= theta <= pi
        assert sintheta <= 1

        LMAX1 = self.LMAX1

        localmemF = cl.LocalMemory(4*LMAX1)
        localmemI = cl.LocalMemory(4*LMAX1)

        ev = self.program.ylm_real(
            queue, (LMAX1,), (LMAX1,),
            np.float32(costheta),
            np.float32(sintheta),
            out.data,
            localmemF,
            localmemI,
        )
        return ev, out

    @optional_out_array(shape=('natwork', 'NLM'), dtype='f4')
    def batch(self, costheta, sintheta, queue=None, out=None):
        """Calculate spherical harmonics for a batch of atoms.

        The calculation is done asynchronously, the returned OpenCL
        event object can be used to wait for the result.
        """
        costheta = np.array(costheta, dtype=np.float32)
        sintheta = np.array(sintheta, dtype=np.float32)
        assert np.shape(costheta) == np.shape(sintheta)
        assert np.all(-1 <= costheta)
        assert np.all(costheta <= 1)
        assert np.all(0 <= sintheta)  # 0 <= theta <= pi
        assert np.all(sintheta <= 1)

        LMAX1 = self.LMAX1
        nAt = np.size(costheta)
        assert nAt > 0
        assert nAt <= self.natwork

        costh_dev = cl_array.to_device(queue, costheta)
        sinth_dev = cl_array.to_device(queue, sintheta)
        localmemF = cl.LocalMemory(4*LMAX1)
        localmemI = cl.LocalMemory(4*LMAX1)

        ev = self.program.ylm_real_batch(
            queue, (LMAX1, nAt), (LMAX1, 1),
            costh_dev.data,
            sinth_dev.data,
            out.data,
            localmemF,
            localmemI,
        )

        return ev, out



"""Tests for the spherical bessel functions."""

# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import assume, given, settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as npst
import numpy as np
import numpy.testing as npt
from scipy.integrate import cumtrapz, quad
import scipy.special

from galumph.sphericalbessel import SphericalBessel


def spherical_bessel(n, z):
    """Calculate the order-n spherical bessel J function of z.

    Uses scipy.special.spherical_jn if available (since scipy 0.18.0),
    otherwise scipy.special.sph_jn (removed in scipy 1.0).
    """
    if hasattr(scipy.special, 'spherical_jn'):
        return scipy.special.spherical_jn(n, z, derivative=False)
    else:
        j, dj = scipy.special.sph_jn(n, z)
        return j[n]


def spherical_bessel_range(nmax, z):
    """Calculate the spherical bessel J function of z for orders up to nmax.

    Uses scipy.special.spherical_jn if available (since scipy 0.18.0),
    otherwise scipy.special.sph_jn (removed in scipy 1.0).
    """
    if hasattr(scipy.special, 'spherical_jn'):
        nrange = np.arange(nmax+1, dtype=int)
        return scipy.special.spherical_jn(nrange, z, derivative=False)
    else:
        j, dj = scipy.special.sph_jn(nmax, z)
        return j


def scipyBessel(LMAX, s, r):
    """Do the same calculation as SphericalBessel.__call__, using scipy."""
    NS, = s.shape
    result = np.empty((LMAX+1, NS))
    for i in range(NS):
        result[:, i] = spherical_bessel_range(LMAX, r*s[i])
    return result


def scipyBesselIntegral(LMAX, s, r):
    """Do the same calculation as SphericalBessel.integral, using scipy."""
    NS, = s.shape
    jl = scipyBessel(LMAX, s, r)
    result = np.zeros_like(jl)
    result[:, 1:] = cumtrapz(jl, x=s[np.newaxis, :]**3, axis=1) * (r**3 / s[1:]**3) / 3
    result[0, 0] = r**3 / 3
    return result


def scipyBesselQuad(L, s, r):
    """Integrate Jl(r*s)*r**2 from 0 to r."""
    def integrand(rr):
        return spherical_bessel(L, s*rr) * rr**2
    result = quad(integrand, 0, r, full_output=True)
    return result[0]


@settings(max_examples=25)
@given(
    LMAX=st.integers(min_value=0, max_value=20),
    worksize=st.integers(min_value=1, max_value=64),
    nchunks=st.integers(min_value=1, max_value=20),
    smax=st.floats(min_value=0.1, max_value=10),
    r=st.floats(min_value=0, max_value=100),
)
def test_single(cl_context, cl_queue,
                LMAX, worksize, nchunks, smax, r):
    """Test calculating the spherical Bessel for a single r value."""
    NS = worksize * nchunks
    assume(NS < 200)
    s = np.linspace(0, smax, num=NS, endpoint=True, dtype='f4')
    clBessel = SphericalBessel(LMAX, s, worksize, context=cl_context)
    ev, jl_dev = clBessel(r, queue=cl_queue)
    jl_sp = scipyBessel(LMAX, s, r)
    ev.wait()
    jl_cl = jl_dev.get(cl_queue)
    tolerance_factor = max((1, smax * r))
    npt.assert_allclose(jl_cl, jl_sp,
                        rtol=(tolerance_factor * 1e-5),
                        atol=(tolerance_factor * 1e-8))


@settings(max_examples=10)
@given(
    LMAX=st.integers(min_value=0, max_value=20),
    worksize=st.integers(min_value=1, max_value=64),
    natwork=st.integers(min_value=2, max_value=16),
    nchunks=st.integers(min_value=1, max_value=20),
    smax=st.floats(min_value=0.1, max_value=10),
    r=npst.arrays(
        dtype=st.sampled_from(('f4', 'f8')),
        shape=st.integers(min_value=1, max_value=16),
        elements=st.floats(min_value=0, max_value=100, width=32),
    ),
)
def test_multiple(cl_context,
                  LMAX, worksize, natwork, nchunks, smax, r):
    """Test calculating the spherical Bessel for multiple r values together."""
    NS = worksize * nchunks
    nAt, = r.shape
    assume(NS * nAt < 1000)
    assume(nAt <= natwork)
    s = np.linspace(0, smax, num=NS, endpoint=True, dtype='f4')
    clBessel = SphericalBessel(LMAX, s, worksize, natwork, context=cl_context)
    jl_dev = clBessel.batch(r)
    jl_cl = jl_dev.get()
    for iAt in range(nAt):
        jl_sp = scipyBessel(LMAX, s, r[iAt])
        tolerance_factor = max((1, smax * r[iAt]))
        npt.assert_allclose(jl_cl[iAt, :, :], jl_sp,
                            rtol=(tolerance_factor * 1e-5),
                            atol=(tolerance_factor * 1e-8))


@settings(max_examples=25)
@given(
    LMAX=st.integers(min_value=0, max_value=20),
    worksize=st.integers(min_value=1, max_value=64),
    nchunks=st.integers(min_value=1, max_value=20),
    smax=st.floats(min_value=0.1, max_value=10),
    r=st.floats(min_value=0, max_value=10),
)
def test_integral_simple(cl_context, cl_queue,
                         LMAX, worksize, nchunks, smax, r):
    """Test the integral kernel against a simple scipy implementation.

    scipyBesselIntegral does the same thing as the integral kernel
    (this checks that the algorithm is correctly implemented)
    """
    NS = worksize * nchunks
    assume(NS < 200)
    s = np.linspace(0, smax, num=NS, endpoint=True, dtype='f4')
    clBessel = SphericalBessel(LMAX, s, worksize, context=cl_context)
    ev, jl_dev = clBessel.integral(r, queue=cl_queue)
    jl_sp = scipyBesselIntegral(LMAX, s, r)
    ev.wait()
    jl_cl = jl_dev.get(cl_queue)
    tolerance_factor = max((1, smax * r))
    npt.assert_allclose(jl_cl, jl_sp,
                        rtol=(tolerance_factor * 1e-4),
                        atol=(tolerance_factor * 1e-6))


@settings(max_examples=10)
@given(
    LMAX=st.integers(min_value=0, max_value=20),
    worksize=st.integers(min_value=16, max_value=64),
    smax=st.floats(min_value=0.1, max_value=10),
    r=st.floats(min_value=0, max_value=10),
    # Don't test all s values as it would take too long, instead take a sample
    # also, don't test the first value, an integral over one step will never be very good
    i_s_test=st.lists(st.integers(min_value=128, max_value=1024),
                      min_size=20, max_size=20, unique=True),
)
def test_integral_quad(cl_context,
                       LMAX, worksize, smax, r, i_s_test):
    """Test the integral kernel against a scipy implementation.

    scipyBesselQuad uses non-uniform sampling to get an accurate
    value for the integral (to check that the algorithm gives
    the correct answer)
    """
    NS = (max(i_s_test) // worksize + 1) * worksize
    assert max(i_s_test) < NS
    assume(NS >= 256)
    s = np.linspace(0, smax, num=NS, endpoint=True, dtype='f4')
    clBessel = SphericalBessel(LMAX, s, worksize, context=cl_context)
    jl_dev = clBessel.integral(r)
    jl_cl = jl_dev.get()

    tolerance_factor = max((1, smax * r))
    for i_s in i_s_test:
        jl_sp = np.array([scipyBesselQuad(L, s[i_s], r)
                          for L in range(LMAX + 1)])
        npt.assert_allclose(jl_cl[:, i_s], jl_sp,
                            rtol=(tolerance_factor * 1e-3),
                            atol=(tolerance_factor * 1e-4))



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import assume, given, settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as npst
import numpy as np
import numpy.testing as npt

from galumph.atomicscattering import AtomicScattering


@st.composite
def alm_parameters(draw,
                   LMAX=st.integers(min_value=3, max_value=20),
                   smax=st.floats(min_value=0.1, max_value=10),
                   worksize=st.integers(min_value=8, max_value=64),
                   natwork=st.integers(min_value=4, max_value=16),
                   ):
    """Hypothesis strategy to get parameters for an Alm array."""
    LMAX = draw(LMAX)
    smax = draw(smax)
    nchunks = draw(st.integers(min_value=4, max_value=16))
    worksize = draw(worksize)
    natwork = draw(natwork)
    NS = worksize * nchunks
    s = np.linspace(0, smax, NS, endpoint=True)
    return dict(
        LMAX=LMAX,
        s=s,
        worksize=worksize,
        natwork=natwork,
    )


@st.composite
def atom_xyzf(draw,
              natoms=st.integers(min_value=1, max_value=5),
              xyz_elements=st.floats(min_value=-10, max_value=10, width=32),
              ff_elements=st.floats(min_value=1, max_value=10, width=32),
              min_distance=0.5,
              # TODO allow specifying s-dependent form factors
              ):
    """Hypothesis strategy to get atomic positions and form factors."""
    natoms = draw(natoms)
    xyz = draw(npst.arrays(
        shape=(natoms, 3),
        elements=xyz_elements,
        dtype=draw(st.sampled_from(('f4', 'f8'))),
    ))
    for i in range(natoms):
        for j in range(i):
            distance = np.linalg.norm(xyz[i, :] - xyz[j, :])
            assume(distance >= min_distance)
    ff = draw(npst.arrays(
        shape=(natoms,),
        elements=ff_elements,
        dtype=draw(st.sampled_from(('f4', 'f8'))),
    ))
    return (xyz, ff)


@settings(max_examples=10)
@given(
    almparams=alm_parameters(),
    xyz_ff=atom_xyzf(natoms=st.integers(min_value=2, max_value=20)),
)
def test_single_multiple(cl_context, almparams, xyz_ff):
    """Check that adding atoms one at a time and all together is the same."""
    kernel = AtomicScattering(context=cl_context, **almparams)
    alm1 = kernel.zeros()
    alm2 = kernel.zeros()
    xyz, ff = xyz_ff
    natoms = xyz.shape[0]

    kernel.add_many_atoms(alm1, xyz, ff)
    for iAt in range(natoms):
        kernel.add_atom(alm2, xyz[iAt, :], ff[iAt])

    np_alm1 = alm1.get()
    np_alm2 = alm2.get()
    npt.assert_allclose(np_alm1, np_alm2, rtol=1e-3, atol=1e-5)


@settings(max_examples=10)
@given(
    almparams=alm_parameters(),
    xyz_ff=atom_xyzf(natoms=st.integers(min_value=2, max_value=20)),
    hyp=st.data(),
)
def test_reorder_atoms(cl_context, almparams, xyz_ff, hyp):
    """Check that Alm is independent of the order in which atoms are added."""
    kernel = AtomicScattering(context=cl_context, **almparams)
    xyz, ff = xyz_ff
    natoms = xyz.shape[0]

    alm1 = kernel(xyz, ff)

    permutation = list(hyp.draw(st.permutations(range(natoms))))
    xyz2 = xyz[permutation, :]
    ff2 = ff[permutation]
    alm2 = kernel(xyz2, ff2)

    np_alm1 = alm1.get()
    np_alm2 = alm2.get()
    npt.assert_allclose(np_alm1, np_alm2, rtol=1e-4, atol=1e-6)



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "European Molecular Biology Laboratory (EMBL)"
__license__ = "LGPLv3+"

from .almarray import AlmArray
from .atomicscattering import AtomicScattering

__all__ = ["AlmArray", "AtomicScattering"]



# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import given
import hypothesis.strategies as st
import pytest

from galumph import gaunt

sympy = pytest.importorskip('sympy')
import sympy.physics.wigner  # noqa: E402

LMAX_TEST = 63


def assert_sympy_equal(a, b):
    """Assert that two sympy expressions are equal."""
    assert sympy.simplify(a - b) == 0


@st.composite
def LMKP(draw):
    L = draw(st.integers(min_value=0, max_value=LMAX_TEST))
    M = draw(st.integers(min_value=0, max_value=L))
    K = draw(st.integers(min_value=M, max_value=L))
    iP = draw(st.integers(min_value=0, max_value=K))
    P = L - K + 2 * iP
    return (L, M, K, P)


def sympy_matrix_element_W(L, M, K, P):
    """Calculate an element of the dlmkp matrix using sympy."""
    wigner0 = sympy.physics.wigner.wigner_3j(L, P, K, 0, 0, 0)
    wignerM = sympy.physics.wigner.wigner_3j(L, P, K, -M, 0, M)
    prefactor = (2*P+1) * sympy.sqrt((2*L+1) * (2*K+1))
    return prefactor * wigner0 * wignerM


def sympy_matrix_element_G(L, M, K, P):
    """Calculate an element of the dlmkp matrix using sympy."""
    standard_gaunt = sympy.physics.wigner.gaunt(L, K, P, M, -M, 0)
    prefactor = sympy.sqrt((2*P+1) * 4 * sympy.pi)
    return prefactor * standard_gaunt


@given(lmkp=LMKP())
def test_sympy_gaunt_vs_wigner(lmkp):
    L, M, K, P = lmkp
    assert_sympy_equal(
        sympy_matrix_element_W(L, M, K, P),
        sympy_matrix_element_G(L, M, K, P),
    )


@given(lmkp=LMKP())
def test_python_vs_sympy_gaunt2(lmkp):
    L, M, K, P = lmkp
    sympy_result = sympy_matrix_element_G(L, M, K, P)
    sympy_sign = sympy.sign(sympy_result)
    python_result, sign = gaunt._modified_gaunt_squared(L, K, P, M)
    if sympy_sign != 0:
        assert sign == sympy_sign
    assert_sympy_equal(
        python_result,
        sympy_result ** 2,
    )


@given(lmkp=LMKP())
def test_python_vs_sympy_gaunt(lmkp):
    L, M, K, P = lmkp
    sympy_result = sympy_matrix_element_G(L, M, K, P)
    sympy_float = float(sympy_result.n(10))
    python_result = gaunt.modified_gaunt(L, K, P, M)
    assert python_result == pytest.approx(sympy_float)



# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import given
import hypothesis.strategies as st
import numpy as np
import numpy.testing as npt

from galumph import almarray


@given(
    cq=st.sampled_from(('context', 'queue')),
    LMAX=st.integers(min_value=0, max_value=63),
    NS=st.integers(min_value=1, max_value=1024),
)
def test_alm_constructor(cl_context, cl_queue, cq, LMAX, NS):
    """Test allocating an Alm array."""
    if cq == 'context':
        cq = cl_context
    else:
        cq = cl_queue
    alm = almarray.AlmArray(cq, LMAX, NS)
    NLM = (LMAX + 1) * (LMAX + 2) // 2
    assert alm.NLM == NLM
    assert alm.shape == (NLM, NS)


@given(
    LMAX=st.integers(min_value=0, max_value=7),
    NS=st.integers(min_value=1, max_value=64),
)
def test_alm_unpack(cl_queue, LMAX, NS):
    """Test the get() function on an AlmArray."""
    alm = almarray.zeros(cl_queue, LMAX, NS)
    # TODO fill with non-uniform values to test that unpacked values are in
    # the correct order
    alm.fill(np.complex64(1-2j))
    npalm = alm.get(unpack=True)
    assert npalm.shape == (LMAX+1, LMAX+1, NS)
    for L in range(LMAX+1):
        for M in range(LMAX+1):
            if M > L:
                npt.assert_array_equal(npalm[L, M, :], 0)
            else:
                npt.assert_array_equal(npalm[L, M, :], 1-2j)



# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import assume, given, settings
import hypothesis.strategies as st
import numpy as np
import numpy.testing as npt
import pyopencl as cl
import pyopencl.array as cl_array
import pytest

from galumph.atomicscattering import AtomicScattering
from galumph.gaunt import shzmat
from galumph.zshift import AlmShiftZ

# Import test strategies from test_atomic_scattering
from test_atomic_scattering import alm_parameters, atom_xyzf


LMAX_TEST = 20


@pytest.fixture(scope='module')
def cl_shzmat(cl_context):
    """Fixture with the dlmkp matrix already copied to the device."""
    dlmkp_host = np.asarray(shzmat(LMAX_TEST), dtype='f4')
    with cl.CommandQueue(cl_context) as queue:
        dlmkp_dev = cl_array.to_device(queue, dlmkp_host)
    return dlmkp_dev


@settings(max_examples=10)
@given(
    almparams=alm_parameters(
        # Need a reasonable LMAX to get accurate results
        LMAX=st.integers(min_value=10, max_value=LMAX_TEST),
        worksize=st.integers(min_value=4, max_value=16),
        smax=st.floats(min_value=0.1, max_value=1),
    ),
    xyz_ff=atom_xyzf(),
    u=st.floats(min_value=0, max_value=5),  # TODO support negative u
)
def test_zshift(cl_context, cl_shzmat, cl_queue, almparams, xyz_ff, u):
    """Compare shifting the Alm versus shifting the atomic positions."""
    s = almparams['s']
    LMAX = almparams['LMAX']
    NSWORK = almparams['worksize']
    # LMAX1 * NSWORK needs to be less than the maximum work group size
    assume((LMAX+1) * NSWORK <= 256)
    assume(u*s[-1] < LMAX/2)
    scattering_kernel = AtomicScattering(context=cl_context, **almparams)
    shift_kernel = AlmShiftZ(context=cl_context,
                             LMAX=LMAX,
                             NSWORK=NSWORK,
                             s=s,
                             dlmkp=cl_shzmat)
    xyz, ff = xyz_ff

    alm_orig = scattering_kernel(xyz, ff)

    ev, alm_almshz = shift_kernel(alm_orig, u, queue=cl_queue)

    xyzshz = xyz + np.array([[0, 0, u]])
    alm_xyzshz = scattering_kernel(xyzshz, ff)

    ev.wait()

    np_alm_almshz = alm_almshz.get(unpack=True)
    # M==0 terms should be real
    npt.assert_array_almost_equal(np_alm_almshz[:, 0, :].imag, 0, decimal=4)

    np_alm_xyzshz = alm_xyzshz.get(unpack=True)
    # Up to LMAX/2 should give good quality
    npt.assert_allclose(np_alm_almshz[:LMAX//2, :LMAX//2, :],
                        np_alm_xyzshz[:LMAX//2, :LMAX//2, :],
                        rtol=1e-2, atol=1e-3 * np.sum(ff))
    npt.assert_allclose(np_alm_almshz, np_alm_xyzshz,
                        rtol=1e-0, atol=1e-1 * np.sum(ff))


@settings(max_examples=10)
@given(
    almparams=alm_parameters(
        # Need a reasonable LMAX to get accurate results
        LMAX=st.integers(min_value=10, max_value=LMAX_TEST),
        worksize=st.integers(min_value=1, max_value=16),
    ),
    xyz_ff=atom_xyzf(),
)
def test_zshift_0(cl_context, cl_shzmat, almparams, xyz_ff):
    """Check that shifting by zero gives exactly the same ALM."""
    s = almparams['s']
    LMAX = almparams['LMAX']
    NSWORK = almparams['worksize']
    # LMAX1 * NSWORK needs to be less than the maximum work group size
    assume((LMAX+1) * NSWORK <= 256)
    scattering_kernel = AtomicScattering(context=cl_context, **almparams)
    shift_kernel = AlmShiftZ(context=cl_context,
                             LMAX=LMAX,
                             NSWORK=NSWORK,
                             s=s,
                             dlmkp=cl_shzmat)
    xyz, ff = xyz_ff
    alm_orig = scattering_kernel(xyz, ff)
    alm_almshz = shift_kernel(alm_orig, 0)

    np_alm_orig = alm_orig.get(unpack=True)
    np_alm_almshz = alm_almshz.get(unpack=True)
    npt.assert_allclose(np_alm_almshz, np_alm_orig)



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2016-2017 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

import os.path

from setuptools import setup


def read(filename):
    repo_dir = os.path.dirname(__file__)
    full_path = os.path.join(repo_dir, filename)
    with open(full_path, 'r') as f:
        return f.read()


setup(
    name='galumph',
    version='0.3.0',
    description='Calculate ALM using GPU acceleration',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://git.embl.de/grp-svergun/galumph',
    author='Christopher Kerr',
    author_email='chris.kerr@mykolab.ch',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    python_requires='>=3.5',
    requires=["pyopencl"],
    packages=[
        'galumph',
    ],
    package_dir={'galumph': 'src/python'},
    package_data={
        'galumph': ["cl-src/*.cl"],
    },
)



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2018-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import assume, given, HealthCheck, settings
import hypothesis.strategies as st
import numpy as np
import numpy.testing as npt
from scipy.stats import linregress

from galumph.atomicscattering import AtomicScattering
from galumph.intensity import Intensity

# Import test strategies from test_atomic_scattering
from test_atomic_scattering import alm_parameters, atom_xyzf


def structure_Guinier(xyz, ff):
    I0 = np.sum(ff)**2

    centre = np.dot(xyz.T, ff) / np.sum(ff)
    r2 = np.sum((xyz - centre[np.newaxis, :])**2, axis=1)
    rG = np.dot(r2, ff) / np.sum(ff)

    return rG, I0


def scattering_Guinier(s, I_s, rGest=None):
    if rGest:
        fitrange = (s <= (1/rGest))
        s = s[fitrange]
        I_s = I_s[fitrange]

    result = linregress(s**2, np.log(I_s))
    rG = -3 * result[0]
    I0 = np.exp(result[1])

    return rG, I0


@settings(max_examples=10)
@given(
    almparams=alm_parameters(),
    xyz_ff1=atom_xyzf(),
    xyz_ff2=atom_xyzf(),
)
def test_cross_intensity(cl_context, cl_queue, almparams, xyz_ff1, xyz_ff2):
    """Test scattering from two separate structures vs. a combined structure.

    The cross scattering plus the self scattering of the two substructures
    should be equal to the self scattering of the combined structure.
    """
    scattering_kernel = AtomicScattering(context=cl_context, **almparams)
    intensity_kernel = Intensity(context=cl_context,
                                 worksize=almparams['worksize'])

    xyz1, ff1 = xyz_ff1
    xyz2, ff2 = xyz_ff2

    ev_alm1, alm1 = scattering_kernel(xyz1, ff1, queue=cl_queue)
    ev_alm2, alm2 = scattering_kernel(xyz2, ff2, queue=cl_queue)

    xyzboth = np.concatenate((xyz1, xyz2), axis=0)
    ffboth = np.concatenate((ff1, ff2), axis=0)
    ev_almboth, almboth = scattering_kernel(xyzboth, ffboth, queue=cl_queue)

    ev_alm1.wait()
    ev1, int1 = intensity_kernel(alm1, queue=cl_queue)
    ev_alm2.wait()
    ev2, int2 = intensity_kernel(alm2, queue=cl_queue)
    evcross, intcross = intensity_kernel.cross(alm1, alm2, queue=cl_queue)

    ev_almboth.wait()
    evboth, intboth = intensity_kernel(almboth, queue=cl_queue)

    cl_queue.finish()
    int1 = int1.get()
    int2 = int2.get()
    intcross = intcross.get()
    intboth = intboth.get()

    npt.assert_allclose(int1 + int2 + intcross, intboth, rtol=1e-4, atol=1e-6)


@settings(max_examples=10,
          # Getting a suitable xyz array requires a lot of filtering
          suppress_health_check=[HealthCheck.filter_too_much])
@given(
    almparams=alm_parameters(
        # need a larger LMAX to get a good fit
        LMAX=st.integers(min_value=20, max_value=30),
        smax=st.floats(min_value=0.3, max_value=3),
        natwork=st.just(16),
    ),
    xyz_ff=atom_xyzf(natoms=st.integers(min_value=2, max_value=20)),
)
def test_Guinier(cl_context, almparams, xyz_ff):
    """Check that the calculated intensity has a sensible Guinier plot.

    Calculates rG and I0 from the Guinier plot and from the atomic structure
    and checks that the two match.
    """
    s = almparams['s']
    scattering_kernel = AtomicScattering(context=cl_context, **almparams)
    intensity_kernel = Intensity(context=cl_context,
                                 worksize=almparams['worksize'])
    xyz, ff = xyz_ff
    # Center the xyz array (not accounting for form factor)
    xyz_mean = np.mean(xyz, axis=0)
    xyz -= xyz_mean[None, :]

    rGxyz, I0xyz = structure_Guinier(xyz, ff)
    # without this, hypothesis gives us structures with all atoms at the origin
    assume(rGxyz * s[-1] > 0.5)
    # or structures where the s step is too big
    assume(rGxyz * s[1] < 0.1)

    alm = scattering_kernel(xyz, ff)
    Icalc_dev = intensity_kernel(alm)
    Icalc = Icalc_dev.get()
    rGalm, I0alm = scattering_Guinier(s, Icalc, rGxyz)

    npt.assert_approx_equal(I0alm, I0xyz, significant=2)
    npt.assert_approx_equal(rGalm, rGxyz, significant=1)



# SPDX-FileCopyrightText: 2018 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl
from pyopencl import array as cl_array

from .util import check_array, PackedLMMixin


class AlmArray(PackedLMMixin, cl_array.Array):
    """Array holding A(L,M,s) representation of scattering amplitudes."""

    def __init__(self, cq, LMAX, NS, dtype='c8', **kwargs):
        """Work out the shape from LMAX and call the superclass constructor."""
        assert LMAX >= 0
        assert NS > 0
        self.LMAX = LMAX
        self.NS = NS
        shape = (self.NLM, self.NS)
        super(AlmArray, self).__init__(cq, shape, dtype=dtype, **kwargs)

    def get(self, queue=None, ary=None, unpack=False, **kwargs):
        """Copy from the device into a numpy array.

        If unpack is True, unpack the array into a lower triangular array
        with dimensions [L,M,S].
        """
        if not unpack:
            return super(AlmArray, self).get(queue=queue, ary=ary, **kwargs)
        unpacked_shape = (self.LMAX1, self.LMAX1, self.NS)
        if ary is None:
            ary = np.zeros(unpacked_shape, dtype=self.dtype)
        else:
            check_array(ary, unpacked_shape)
        for L in range(self.LMAX1):
            packed_range = slice(self.indexLM(L, 0), self.indexLM(L+1, 0))
            ary[L, 0:L+1, :] = self[packed_range, :].get(queue=queue, **kwargs)
        return ary


def empty(cq, LMAX, NS, dtype='c8', **kwargs):
    """Create an AlmArray without initializing the memory."""
    return AlmArray(cq, LMAX, NS, dtype=dtype, **kwargs)


def zeros(cq, LMAX, NS, dtype='c8', **kwargs):
    """Create an AlmArray initialized to zero."""
    alm = AlmArray(cq, LMAX, NS, dtype=dtype, **kwargs)
    if isinstance(cq, cl.CommandQueue):
        alm.fill(0, queue=cq)
    else:
        with cl.CommandQueue(cq) as queue:
            alm.fill(0, queue)
    return alm



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2017-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "European Molecular Biology Laboratory (EMBL)"
__license__ = "LGPLv3+"

import numpy as np
import pyopencl as cl
from pyopencl import array as cl_array

from . import almarray
from .sphericalbessel import SphericalBessel
from .sphericalharmonic import SphericalHarmonic
from .util import check_array, ClProgram, PackedLMMixin, with_some_queue


def sicopolar(xyz):
    x = xyz[0]
    y = xyz[1]
    z = xyz[2]
    rho = np.linalg.norm(xyz[:2])
    if rho > 0:
        cosphi = x/rho
        sinphi = y/rho
    else:
        cosphi = 0
        sinphi = 0
    r = np.linalg.norm(xyz)
    if r > 0:
        costheta = z/r
        sintheta = rho/r
    else:
        costheta = 0
        sintheta = 0
    return r, costheta, sintheta, complex(cosphi, sinphi)


def sicopolar_array(xyz):
    x = xyz[:, 0]
    y = xyz[:, 1]
    z = xyz[:, 2]

    rho = np.linalg.norm(xyz[:, :2], axis=1)
    rho_gt0 = (rho > 0)
    eiphi = x + 1j * y
    eiphi[rho_gt0] /= rho[rho_gt0]

    r = np.linalg.norm(xyz, axis=1)
    r_gt0 = (r > 0)
    costheta = np.zeros_like(r)
    sintheta = np.zeros_like(r)
    costheta[r_gt0] = z[r_gt0]/r[r_gt0]
    sintheta[r_gt0] = rho[r_gt0]/r[r_gt0]
    return r, costheta, sintheta, eiphi


class AtomicScattering(PackedLMMixin, ClProgram):
    """Kernel for calculating the ALM scattering from an atomic structure."""

    def __init__(self, LMAX, s,
                 worksize=64,
                 natwork=16,
                 context=None):
        assert LMAX >= 0
        self.LMAX = LMAX
        s = np.asanyarray(s)
        NS, = s.shape
        assert NS > 0
        assert np.all(s >= 0)
        self.NS = NS
        self.s = s
        assert worksize > 0
        assert NS % worksize == 0
        self.worksize = worksize
        assert natwork > 0
        self.natwork = natwork
        super(AtomicScattering, self).__init__(
            filename='alm-add-atom.cl',
            options=["-DNATWORK=%d" % natwork, "-DWORKSIZE=%d" % worksize],
            context=context)
        context = self.context

        self.sphbes = SphericalBessel(LMAX, s, worksize, natwork, context=context)
        self.sphharm = SphericalHarmonic(LMAX, natwork, context=context)

    def zeros(self):
        """Return an AlmArray with the correct shape initialized to zero."""
        with cl.CommandQueue(self.context) as queue:
            return almarray.zeros(queue, self.LMAX, self.NS)

    @with_some_queue
    def add_atom(self, alm, xyz, ff, queue=None):
        """Add the scattering from a single atom to the ALM.

        Args:
            alm: AlmArray
            xyz: Cartesian coordinates of the atom
            ff: Atomic form factor: single value or an array of length NS

        """
        assert alm.context == self.context
        check_array(alm, (self.NLM, self.NS), dtype='c8')
        assert queue is not None
        LMAX1 = self.LMAX1
        NS = self.NS
        worksize = self.worksize

        ff = np.asanyarray(ff)
        if np.isscalar(ff) or (np.shape(ff) == ()):
            ff = np.ones_like(self.s) * ff
        else:
            assert ff.shape == self.s.shape

        r, costheta, sintheta, eiphi = sicopolar(xyz)

        evY, ylm_dev = self.sphharm(costheta, sintheta, queue=queue)
        evJ, jl_dev = self.sphbes(r, queue=queue)

        ff32 = np.array(ff, dtype=np.float32)
        ff_dev = cl_array.to_device(queue, ff32)

        ev = self.program.alm_add_atom(queue,
                                       (LMAX1, NS),
                                       (1, worksize),
                                       np.complex64(eiphi),
                                       ylm_dev.data,
                                       jl_dev.data,
                                       ff_dev.data,
                                       alm.data,
                                       wait_for=(evY, evJ))
        return ev, alm

    @with_some_queue
    def add_many_atoms(self, alm, xyz, ff, queue=None):
        """Add the scattering from an array of atoms to the ALM.

        Args:
            alm: AlmArray
            xyz: Cartesian coordinates of the atoms
            ff: Atomic form factors: shape either (natoms) or (natoms, NS)

        """
        assert alm.context == self.context
        check_array(alm, (self.NLM, self.NS), dtype='c8')
        assert queue is not None
        LMAX1 = self.LMAX1
        NS = self.NS
        natwork = self.natwork
        worksize = self.worksize

        natoms, ndim = np.shape(xyz)
        assert natoms > 0
        assert ndim == 3

        ff = np.asanyarray(ff)
        if np.ndim(ff) == 1:
            ff = np.ones_like(self.s) * ff[:, np.newaxis]
        check_array(ff, (natoms, NS))

        ylm_local = cl.LocalMemory(8*natwork*LMAX1)

        for iAt0 in range(0, natoms, natwork):
            iAtN = min(natoms, iAt0+natwork)

            r, costheta, sintheta, eiphi = sicopolar_array(xyz[iAt0:iAtN, :])

            evY, ylm_dev = self.sphharm.batch(costheta, sintheta, queue=queue)
            evJ, jl_dev = self.sphbes.batch(r, queue=queue)

            ff32 = np.array(ff[iAt0:iAtN, :], dtype=np.float32)
            ff_dev = cl_array.to_device(queue, ff32)
            eiphi32 = np.array(eiphi, dtype=np.complex64)
            eiphi_dev = cl_array.to_device(queue, eiphi32)

            ev = self.program.alm_add_many_atoms(queue,
                                                 (LMAX1, NS),
                                                 (1, worksize),
                                                 np.uint32(iAtN - iAt0),
                                                 eiphi_dev.data,
                                                 ylm_dev.data,
                                                 jl_dev.data,
                                                 ff_dev.data,
                                                 alm.data,
                                                 ylm_local,
                                                 wait_for=(evY, evJ))
            ev.wait()
        return ev, alm

    @with_some_queue
    def __call__(self, xyz, ff, queue=None):
        """Calculate the atomic scattering from the given atoms.

        Allocates an AtomAlm array, zeros it, then adds the scattering from
        the atoms given the (x,y.z) positions and the atomic form factors.
        """
        alm = self.zeros()
        ev, alm = self.add_many_atoms(alm, xyz, ff, queue=queue)
        return ev, alm



"""Pure Python implementation of Gaunt coefficients.

Gaunt coefficients give the spatial integral of the product of three
spherical harmonic functions. In GALUMPH, they are used in the ALM
translation function.

The code here is based on Formula 4.2 of:
  J. Rasch and A. C. H. Yu, 'Efficient Storage Scheme for
  Pre-calculated Wigner 3j, 6j and Gaunt Coefficients', SIAM
  J. Sci. Comput. Volume 25, Issue 4, pp. 1416-1428 (2003)
  https://doi.org/10.1137/S1064827503422932
However, the formula in that article has some errors, which are corrected here.

For a translation along the Z axis, we only need values with:
    m_3 == 0, m_1 == -m_2
This allows simplifying the formula and referring to m_1 as plain `m`.
"""

# SPDX-FileCopyrightText: 2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

__copyright__ = "Christopher Kerr"
__license__ = "LGPLv3+"

from fractions import Fraction
import functools
import math
import operator


def prod(iterable, *, start=1):
    """Product of a sequence of numbers.

    Reimplementation of math.prod, which is only available for Python >= 3.8.
    """
    return functools.reduce(operator.mul, iterable, start)


def _calculate_factorials(N):
    """Return a tuple containing the factorials of all numbers from 0 to N.

    Since Python integers are arbitrary-precision, the values are exact.
    """
    assert N >= 0
    # Special case for 0
    fac_i = 1
    fac_list = [fac_i]
    for i in range(1, N+1):
        fac_i *= i
        fac_list.append(fac_i)
    return tuple(fac_list)


def delta_squared(l_1, l_2, l_3, *, factorials=None):
    """Square of the Delta function defined in Formula 2.3 of Rasch2003."""
    if factorials is None:
        factorials = _calculate_factorials(l_1 + l_2 + l_3 + 1)
    return Fraction(
        prod((
            factorials[l_1 + l_2 - l_3],
            factorials[l_2 + l_3 - l_1],
            factorials[l_3 + l_1 - l_2],
        )),
        factorials[l_1 + l_2 + l_3 + 1],
    )


def _gaunt_ksum(l_1, l_2, l_3, m, *, factorials):
    """Part of the Gaunt coefficient function involving a sum over k."""
    k_min = max(l_1 - m - l_3, l_2 - m - l_3, 0)
    k_max = min(l_1 - m, l_2 - m, l_1 + l_2 - l_3)

    def kpart(k):
        return Fraction(
            (-1) ** k,
            prod((
                factorials[k],
                factorials[k + l_3 + m - l_1],
                factorials[k + l_3 + m - l_2],
                factorials[l_1 + l_2 - l_3 - k],
                factorials[l_1 - m - k],
                factorials[l_2 - m - k],
            )),
        )

    return sum(kpart(k) for k in range(k_min, k_max + 1))


def _gaunt_sqrt_arg(l_1, l_2, l_3, m, *, factorials):
    """Calculate the part of the formula that needs to be square-rooted."""
    return prod((
        (2 * l_1 + 1),
        (2 * l_2 + 1),
        (2 * l_3 + 1),
        factorials[l_1 + m],
        factorials[l_1 - m],
        factorials[l_2 + m],
        factorials[l_2 - m],
        factorials[l_3 + 0],
        factorials[l_3 - 0],
    ))


def _gaunt_big_L_part(l_1, l_2, l_3, *, factorials):
    """The part of the Gaunt coefficient formula involving the capital L."""
    assert (l_1 + l_2 + l_3) % 2 == 0
    L = (l_1 + l_2 + l_3) // 2
    return Fraction(
        factorials[L],
        prod((
            factorials[L - l_1],
            factorials[L - l_2],
            factorials[L - l_3],
        )),
    )


def _modified_gaunt_squared(l_1, l_2, l_3, m, *, factorials=None):
    """Square magnitude and sign of the modified Gaunt coefficient.

    Here we are calculating a modified Gaunt coefficient which differs from
    the standard coefficient by a factor of sqrt((2 * l_3 + 1) / 4 * pi). That
    modification is applied here.
    """
    if factorials is None:
        factorials = _calculate_factorials(l_1 + l_2 + l_3 + 1)
    non_sqrt_part = prod((
        delta_squared(l_1, l_2, l_3, factorials=factorials),
        _gaunt_ksum(l_1, l_2, l_3, m, factorials=factorials),
        _gaunt_big_L_part(l_1, l_2, l_3, factorials=factorials),
    ))
    sq_mag = prod((
        _gaunt_sqrt_arg(l_1, l_2, l_3, m, factorials=factorials),
        non_sqrt_part * non_sqrt_part,
        (2 * l_3 + 1),  # For standard Gaunt, omit this
    ))  # For standard Gaunt, divide by (4 * math.pi)
    sign = (-1) ** ((l_1 + l_2 - l_3) // 2)
    if non_sqrt_part < 0:
        sign *= -1
    return sq_mag, sign


def modified_gaunt(l_1, l_2, l_3, m, *, factorials=None):
    """Modified Gaunt coefficient used in the Z translation matrix."""
    sq_mag, sign = _modified_gaunt_squared(l_1, l_2, l_3, m, factorials=factorials)
    return sign * math.sqrt(sq_mag)


def shzmat(LMAX):
    """Calculate the array of Gaunt coefficients used for ALM Z translation."""
    factorials = _calculate_factorials(4 * LMAX + 1)
    dlmkp = list()
    for L in range(LMAX + 1):
        for M in range(L + 1):
            for K in range(M, L + 1):
                for P in range(L - K, L + K + 1, 2):
                    dlmkp.append(modified_gaunt(L, K, P, M, factorials=factorials))
    return dlmkp



# SPDX-FileCopyrightText: 2016 European Molecular Biology Laboratory (EMBL)
# SPDX-FileCopyrightText: 2017-2020 Christopher Kerr
#
# SPDX-License-Identifier: LGPL-3.0-or-later

from hypothesis import assume, example, given, settings
import hypothesis.strategies as st
import hypothesis.extra.numpy as npst
import numpy as np
import numpy.testing as npt
from scipy.special import sph_harm

from galumph.sphericalharmonic import SphericalHarmonic


def scipy_SphHarm(LMAX, theta):
    NLM = (LMAX+1)*(LMAX+2)//2
    ylm = np.empty((NLM,), dtype=np.complex128)
    for L in range(LMAX+1):
        il0 = L*(L+1)//2
        ilend = il0+L+1
        ylm[il0:ilend] = sph_harm(np.arange(L+1), L, 0, theta)
    return ylm.real


@settings(max_examples=25)
@given(
    LMAX=st.integers(min_value=3, max_value=20),
    theta=st.floats(min_value=0, max_value=np.pi),
)
@example(LMAX=3, theta=(np.pi/2))  # Issue #24
def test_single(cl_context, cl_queue, LMAX, theta):
    """Test calculating the spherical harmonics for one (theta, phi) pair."""
    clSphHarm = SphericalHarmonic(LMAX, context=cl_context)
    costheta = np.cos(theta)
    sintheta = np.sin(theta)
    # Rounding error sometimes gives theta slightly greater than pi
    assume(sintheta >= 0)

    ev, ylm_dev = clSphHarm(costheta, sintheta, queue=cl_queue)
    ylm_sp = scipy_SphHarm(LMAX, theta)
    ev.wait()
    ylm_cl = ylm_dev.get(cl_queue)

    npt.assert_allclose(ylm_cl[:], ylm_sp, rtol=1e-4, atol=1e-6)


@settings(max_examples=10)
@given(
    LMAX=st.integers(min_value=3, max_value=20),
    theta=npst.arrays(
        shape=st.tuples(
            st.integers(min_value=1, max_value=16),
        ),
        elements=st.floats(min_value=0, max_value=np.float32(np.pi), width=32),
        dtype=st.sampled_from(('f4', 'f8')),
    ),
)
def test_multiple(cl_context, cl_queue, LMAX, theta):
    """Test calculating the spherical harmonics for arrays of theta."""
    clSphHarm = SphericalHarmonic(LMAX, context=cl_context)
    nAt = theta.shape[0]
    costheta = np.cos(theta)
    sintheta = np.sin(theta)
    # Rounding error sometimes gives theta slightly greater than pi
    assume(np.all(sintheta >= 0))

    ev, ylm_dev = clSphHarm.batch(costheta, sintheta, queue=cl_queue)

    NLM = (LMAX+1)*(LMAX+2)//2
    ylm_sp = np.empty((nAt, NLM), dtype=np.complex128)
    for iAt in range(nAt):
        ylm_sp[iAt, :] = scipy_SphHarm(LMAX, theta[iAt])
    ev.wait()
    ylm_cl = ylm_dev.get(cl_queue)

    npt.assert_allclose(ylm_cl[:nAt, :], ylm_sp, rtol=1e-4, atol=1e-6)
